import abc
from enum import Enum
from cifrador.cifrador import CifradorSimetrico, CifradorAsimetrico
from descifrador.descifrador import DescifradorSimetrico, DescifradorAsimetrico
from comunicaciones.comunicaciones import FachadaComunicaciones
from comunicaciones.notificador import IObserver     
   
# Enumeración opciones   
class Opcion(Enum):
    SIMETRICO = 1
    HIBRIDO = 2

# Enumeración rango del usuario avanzado
class Rango(Enum):
    APRENDIZ  = 1
    COMPAÑERO = 2
    MAESTRO   = 3

# Interfaz IMenuCifrado
class IMenuCifrado(metaclass = abc.ABCMeta):
    @abc.abstractmethod
    def set_opcion(self, opcion):
        pass   

# Clase usuario normal. Implementa IMenuCifrado e IObserver
class Usuario(IMenuCifrado, IObserver):
    def __init__(self, nombre, login, password):
        self.__simetrico = None
        self.__asimetrico = None
        self.__numero_accesos = 0
        self.__nombre = nombre
        self._login = login
        self._password = password 
        self._conexion = FachadaComunicaciones()
            
    # Envía datos cifrados    
    def __envia_datos(self, msg, clave, nonce, tag):
        try:
            self._conexion.conecta()
        except:
            print("Error de conexion")  
            return    
        try:
            self._conexion.envia_msg(msg)
            self._conexion.envia_clave(clave)
            self._conexion.envia_nonce(nonce)
            self._conexion.envia_tag(tag)
        except:
            print("Error enviando datos por socket")    
        finally:    
            self._conexion.desconecta()  
    
    
    def _introduce_mensaje(self):
        return input("Introduzca el mensaje a cifrar: ")
        
    # Genera clave simétrica de 16 bytes AES automáticamente
    def _introduce_clave(self):
        clave = CifradorSimetrico.genera_clave()
        print (f"Introduzca la clave de cifrado: {clave}")
        return clave    
        
    # Cifra con algoritmo simétrico AES
    def cifra_simetrico(self, msg, clave):
        if self.__simetrico == None:
               print("Creando cifrador simétrico...")
               self.__simetrico = CifradorSimetrico()
        
        self.__numero_accesos = self.__numero_accesos + 1
        self.__simetrico.cifra(msg, clave)
        msg, nonce, tag = self.__simetrico.get_msg()
        
        print (f"Mensaje cifrado= {msg}")
        self.__envia_datos(msg, clave, nonce, tag)
       
        
    # Cifra con algoritmo híbrido AES-RSA
    def cifra_hibrido(self, msg):
        if self.__asimetrico == None:
            print("Creando cifrador asimétrico...")
            self.__asimetrico = CifradorAsimetrico()
            if self.__simetrico == None:
                print("Creando cifrador simétrico...")
                self.__simetrico = CifradorSimetrico()
            print("Por favor, espere...")
            self.__clave_sesion = self.__simetrico.genera_clave()
            self.__asimetrico.cifra(self.__clave_sesion)
            self.clave_sesion_cifrada = self.__asimetrico.get_msg()   
        
        self.__numero_accesos = self.__numero_accesos + 1
        self.__simetrico.cifra(msg,self.__clave_sesion)
        msg, nonce, tag = self.__simetrico.get_msg() 
        
        print (f"Mensaje cifrado= {msg}")
        self.__envia_datos(msg, self.clave_sesion_cifrada, nonce, tag)
        
    # Define opción
    def set_opcion(self, opcion):
        if opcion == Opcion.SIMETRICO:   
            msg = self._introduce_mensaje()
            clave = self._introduce_clave()
            self.cifra_simetrico(msg, clave)
        elif opcion == Opcion.HIBRIDO:
            msg = self._introduce_mensaje()
            self.cifra_hibrido(msg)   
             
 # Clase para el usuario avanzado. Hereda de usuario normal con restricciones.            
class UsuarioAvanzado(Usuario):
    def __init__(self, nombre, login, password, simbolico, filtrar):
        super().__init__(nombre, login, password)
        self.__rango = Rango.COMPAÑERO
        self.__simbolico = simbolico
        self.__filtrar = filtrar
        self._conexion.adjunta(self) # Adjunta observador
        CifradorAsimetrico().crea_claves()
        
    # Activa servidor
    def espera_mensaje_cifrado(self):
        self._conexion.activa_servidor()
    
    # Filta spam en mensaje descifrado
    def _filtra_mensaje(self, mensaje_descifrado):
        print("Spam filtrado")
    
    # Parte del patrón Observer. Notificación.    
    def notifica(self, mensaje_cifrado, clave, nonce, tag):
        print("Recibiendo mensaje...")
        print("Probando estrategia simétrica AES...")
        
        # Prueba con estrategia simétrica AES
        try:
            descifrador_simetrico = DescifradorSimetrico()
            descifrador_simetrico.descifra(mensaje_cifrado, clave, nonce, tag)
            self.mensaje_descifrado = descifrador_simetrico.get_msg()
            print(f"Posible mensaje descifrado: {self.mensaje_descifrado}")
                   
        except:
            print("mensaje no cifrado con algoritmo simétrico AES");
        
        
        print("Probando estrategia híbrida AES-RSA...")
        
        # Prueba con estrategia asimétrica AES-RSA
        try:
            descifrador_asimetrico = DescifradorAsimetrico()
            descifrador_asimetrico.descifra(clave)    
            clave_descifrada = descifrador_asimetrico.get_msg()
            descifrador_simetrico.descifra(mensaje_cifrado, clave_descifrada, nonce, tag)
            self.mensaje_descifrado = descifrador_simetrico.get_msg()
            print(f"Posible mensaje descifrado: {self.mensaje_descifrado}")
        
        except:
           print("mensaje no cifrado con algoritmo híbrido AES-RSA")
           
        # Comprueba que requiere filtrado spam
        if self.__filtrar == "s":
            self._filtra_mensaje(self.mensaje_descifrado) 